#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include "smartio.h"

#ifdef DEBUG
#define LIBSIO_DEBUG 1
#else
#define LIBSIO_DEBUG 0
#endif

// Local function prototypes -------------------------------

static int OpenSIODevice(const char *dev);
static int ReadDeviceInfo(t_sio_ver *pInfo);
static void InitSIOPort(const char *dev, char val);
static void WriteSIOPort(const char *dev, char val);
static char ReadSIOPort(const char *dev);


// Library function definitions ----------------------------

// Smart I/O device information -----------------------

char SIOReadDeviceVersion(void)
{
  t_sio_ver SioInfo;
  char rtn = '0';

  if (ReadDeviceInfo(&SioInfo)) {
    rtn = (char)SioInfo.DeviceVer;
  }
	
  return rtn;
}

unsigned short SIOReadDevicePartType(void)
{
  t_sio_ver SioInfo;
  int rtn = 0;

  if (ReadDeviceInfo(&SioInfo)) {
    rtn = (unsigned short)SioInfo.DeviceType;
  }
	
  return rtn;
}

unsigned short SIOReadFirmwareVersion(void)
{
  t_sio_ver SioInfo;
  int rtn = 0;

  if (ReadDeviceInfo(&SioInfo)) {
    rtn = (unsigned short)SioInfo.FirmwareLevel;
  }
	
  return rtn;
}

// Back Light -----------------------------------------

void SIOBackLightOn(void)
{
  int fd = OpenSIODevice(DEV_BACKLITE);
  if (0 <= fd) {
    char val = 1;
    ioctl(fd, SMARTIO_BL_CONTROL, &val);
    close(fd);
  }
}

void SIOBackLightOff(void)
{
  int fd = OpenSIODevice(DEV_BACKLITE);
  if (0 <= fd) {
    char val = 0;
    ioctl(fd, SMARTIO_BL_CONTROL, &val);
    close(fd);
  }
}

void SIOSetBackLightPwm(char val)
{
  int fd = OpenSIODevice(DEV_BACKLITE);
  if (0 <= fd) {
    // SMARTIO_BL_CONTRAST is really brightness
    ioctl(fd, SMARTIO_BL_CONTRAST, &val);
    close(fd);
  }
}

// Passive matrix screen contrast
void SIOSetVEEPwm(char val)
{
  WriteSIOPort(DEV_VEE, val);
}
/*
#define	SMARTIO_DUMMY_VALUE	_IOW(SMARTIO_IOCTL_BASES, 0, int)
{
  int fd = OpenSIODevice(DEV_VEE);
  if (0 <= fd) {
    // There's only one control for this device
    // It hasn't been defined so use SMARTIO_DUMMY_VALUE
    ioctl(fd, SMARTIO_DUMMY_VALUE, &val);
    close(fd);
  }
}
*/

// Smart I/O option -----------------------------------

void SIOSelectOption(char option)
{
  int fd = OpenSIODevice(DEV_SELOPT);
  if (0 <= fd) {
#if LIBSIO_DEBUG
    int result = write(fd, &option, 1);
    if (0 > result) {
      printf("SIO select option error (%d)\n", result);
    }
#else
    write(fd, &option, 1);
#endif
    close(fd);
  }
}

// Keypad ---------------------------------------------

#define MAX_COLS 8
#define MAX_ROWS 8

// 0-8, 0-8 are valid inputs (0-4, 0-5 on Bitsy)
char SIOSetKeypadSize(char cols, char rows)
{
  char rtn = (char)-1;

  int fd = OpenSIODevice(DEV_KEYPAD);
  if (0 <= fd) {
    char cr[2] = {0, 0};
    if (cols > 0 && cols <= MAX_COLS && rows > 0 && rows < MAX_ROWS) {
      cr[0] = cols;
      cr[1] = rows;
      rtn = 0;
    }
    else if (cols == 0 && rows == 0) {
      rtn = 0;
    }
    ioctl(fd, SMARTIO_KPD_SETUP, &cr);
    close(fd);
  }

  return rtn;
}

// use -1 to wait for key press
int SetKeypadTimeout(long time)
{
  int rtn = -1;

  if (-1 <= time) {
    int fd = OpenSIODevice(DEV_KEYPAD);
    if (0 <= fd) {
      ioctl(fd, SMARTIO_KPD_TIMEOUT, &time);
      close(fd);
      rtn = 0;
    }
  }

  return rtn;
}

int SIOReadKeypad(void)
{
  char rtn = (char)-1;
  int fd = OpenSIODevice(DEV_KEYPAD);
  if (0 <= fd) {
#if LIBSIO_DEBUG
    int result = read(fd, &rtn, sizeof(rtn));
    if (0 > result) {
      printf("SIO read keypad error (%d)\n", result);
    }
#else
    read(fd, &rtn, sizeof(rtn));
#endif
    close(fd);
  }
  return (int)rtn;
}

// A/D ------------------------------------------------

// Option 2
int SIOAnalogConversion(char chan)
{
  short rtn = -1;

  int fd = OpenSIODevice(DEV_ADC);
  if (0 <= fd) {
    rtn = chan;
#if LIBSIO_DEBUG
    int result = read(fd, &rtn, sizeof(rtn));
    if (0 > result) {
      printf("SIO read %s error (%d)\n", DEV_ADC, result);
    }
#else
    read(fd, &rtn, sizeof(rtn));
#endif
    close(fd);
  }

  return (int)rtn;
}

// Digital I/O ----------------------------------------

void SIOInitDDRA(char val)
{
  // Options 1 & 3
  InitSIOPort(DEV_PORTA, val);
}

void SIOInitDDRB(char val)
{
  InitSIOPort(DEV_PORTB, val);
}

void SIOInitDDRC(char val)
{
  InitSIOPort(DEV_PORTC, val);
}

void SIOInitDDRD(char val)
{
  InitSIOPort(DEV_PORTD, val);
}

void SIOWritePortA(char val)
{
  // Options 1 & 3
  WriteSIOPort(DEV_PORTA, val);
}

void SIOWritePortB(char val)
{
  WriteSIOPort(DEV_PORTB, val);
}

void SIOWritePortC(char val)
{
  WriteSIOPort(DEV_PORTC, val);
}

void SIOWritePortD(char val)
{
  WriteSIOPort(DEV_PORTD, val);
}

char SIOReadPortA(void)
{
  // Options 1 & 3
  return ReadSIOPort(DEV_PORTA);
}

char SIOReadPortB(void)
{
  return ReadSIOPort(DEV_PORTB);
}

char SIOReadPortC(void)
{
  return ReadSIOPort(DEV_PORTC);
}

char SIOReadPortD(void)
{
  return ReadSIOPort(DEV_PORTD);
}

// UCB1200 I/O ----------------------------------------

void SIOInitUCB(unsigned short val)
{
  int fd = OpenSIODevice(DEV_UCB_IO);
  if (0 <= fd) {
    // There's only one control for this device
    // It hasn't been defined so use SMARTIO_DUMMY_VALUE
    ioctl(fd, SMARTIO_DUMMY_VALUE, &val);
    close(fd);
  }
}

void SIOWriteUCB(unsigned short val)
{
  int fd = OpenSIODevice(DEV_UCB_IO);
  if (0 <= fd) {
#if LIBSIO_DEBUG
    int result = write(fd, &val, sizeof(val));
    if (0 >= result) {
      printf("SIO write %s error (%d)\n", DEV_UCB_IO, result);
    }
#else
    write(fd, &val, sizeof(val));
#endif
    close(fd);
  }
}

unsigned short SIOReadUCBDigital(void)
{
  unsigned short rtn = ~0;

  int fd = OpenSIODevice(DEV_UCB_IO);
  if (0 > fd) {
    return rtn;
  }

  int result = read(fd, &rtn, sizeof(rtn));
  if (0 > result) {
#if LIBSIO_DEBUG
    printf("SIO read %s error (%d)\n", DEV_UCB_IO, result);
#endif
    rtn = ~0;
  }
  close(fd);

  return rtn;
}

#define MAX_UCB_ADC 4
const char *dev_ucb_adc[MAX_UCB_ADC] = {
  DEV_UCB_ADC0,
  DEV_UCB_ADC1,
  DEV_UCB_ADC2,
  DEV_UCB_ADC3
};

unsigned short SIOReadUCBAnalog(unsigned char chan)
{
  unsigned short rtn = ~0;

  if (MAX_UCB_ADC <= chan) {
    return rtn;
  }

  int fd = OpenSIODevice(dev_ucb_adc[chan]);
  if (0 > fd) {
    return rtn;
  }

  int result = read(fd, &rtn, sizeof(rtn));
  if (0 > result) {
#if LIBSIO_DEBUG
    printf("SIO read %s error (%d)\n", dev_ucb_adc[chan], result);
#endif
    rtn = ~0;
  }
  close(fd);

  return rtn;
}


// Local function definitions ------------------------------

static int OpenSIODevice(const char *dev)
{
  int fd = open(dev, O_RDWR);

#if LIBSIO_DEBUG
  if (0 > fd) {
    printf("SIO open device %s error (%d)\n", dev, fd);
  }
#endif

  return fd;
}
 
static int ReadDeviceInfo(t_sio_ver *pInfo)
{
  int rtn = 0;

  if (pInfo) {
    int fd = OpenSIODevice(DEV_VER);
    if (0 <= fd) {
      int result = read(fd, pInfo, sizeof(t_sio_ver));
      if (0 <= result) {
	rtn = 1;
      }
      close(fd);
    }
  }

  return rtn;
}
 
static void WriteSIOPort(const char *dev, char val)
{
  int fd = OpenSIODevice(dev);
  if (0 <= fd) {
#if LIBSIO_DEBUG
    int result = write(fd, &val, sizeof(val));
    if (0 > result) {
      printf("SIO write %s error (%d)\n", dev, result);
    }
#else
    write(fd, &val, sizeof(val));
#endif
    close(fd);
  }
}

static void InitSIOPort(const char *dev, char val)
{
  int fd = OpenSIODevice(dev);
  if (0 <= fd) {
    ioctl(fd, SMARTIO_PORT_CONFIG, &val);
    close(fd);
  }
}

static char ReadSIOPort(const char *dev)
{
  char rtn = 0;

  int fd = OpenSIODevice(dev);
  if (0 <= fd) {
#if LIBSIO_DEBUG
    int result = read(fd, &rtn, sizeof(rtn));
    if (0 > result) {
      printf("SIO read %s error (%d)\n", dev, result);
    }
#else
    read(fd, &rtn, sizeof(rtn));
#endif
    close(fd);
  }

  return rtn;
}
